#!/usr/bin/perl -w
use strict;

my @TargMatrix;
my (@names, @emails, @colls, @addrs, @pseus, @wwwws);
my $numplayers;
my $numtargs = 3;
my ($x, $y);

# These are some magic numbers. Don't change them unless you are Edward M Nokes,
# the executor of his estate, or a mandated UN representative.
my $NOKES1 = "0001900";
my $NOKES2 = "0062000";
my $NOKES3 = "0131900";

# contains($i, @array): A simple routine to check if an array contains a given
# value
sub contains {
	my $int = shift;
	my @array = @_;
	my $temp = 0;
	foreach (@array) {
		$temp = 1 if ($_ == $int);
		last if $temp;
	}
	return $temp;
}

# floyd(\@AdjacencyMatrix): Floyd's algorithm. Operates in O(n^3), returns a 
# matrix of pairwise distances.
sub floyd {
	my $aref = shift;
	my @adjacency = @$aref;
	my ($i, $j, $k);
	my @distance;
	for $i (1..$numplayers) {
		for $j (1..$numplayers) {
			$distance[$i][$j] = $numplayers + 1;
			$distance[$i][$j] = 1 if $adjacency[$i][$j];
			$distance[$i][$j] = 0 if $i == $j;
		}
	}
		
	for $k (1..$numplayers) {
		for $i (1..$numplayers) {
			for $j (1..$i) {
				my $through_k = $distance[$i][$k] + $distance[$k][$j];
				if ($through_k < $distance[$i][$j]) {
					$distance[$i][$j] = $through_k;
					$distance[$j][$i] = $through_k;
				}
			}
		}
	}
	return @distance;
}

# reallocate(\@TargettingMatrix): Reallocation algorithm. Probabilistic, so may
# take forever. Rarely does, unless there's a problem. Need to write something 
# to deal with that case. Also, solve Halting Problem.	
sub reallocate {
	my $matref = shift;
	my @OrigMatrix = @$matref;
	my @NewMatrix;
	my @ntargs;
	my @nasses;
	my ($i, $j);
	my $fail = 1;
	while ($fail) {
		for $i (1..$numplayers) {
			$ntargs[$i] = 0;
			$nasses[$i] = 0;
			for $j (1..$numplayers) {
				$NewMatrix[$i][$j] = $OrigMatrix[$i][$j];
				$ntargs[$i]++ if $NewMatrix[$i][$j] == 1;
				$nasses[$i]++ if $NewMatrix[$i][$j] == -1;
			}
		}
		$fail = 0;
		for $i (1..$numplayers) {
			my $count = 0;
			while (($ntargs[$i] != $numtargs) && ($count < $numplayers * 2)) {
				$j = int(rand($numplayers)) + 1;
				if (($j != $i) && ($nasses[$j] != $numtargs) && ($NewMatrix[$i][$j] == 0)) {
					$ntargs[$i]++;
					$nasses[$j]++;
					$NewMatrix[$i][$j] = 1;
					$NewMatrix[$j][$i] = -1;
					$count = 0;
				}
				$count++;
			}
			$fail++ if $count == $numplayers * 2;
		}
	}
	return @NewMatrix;
}

# girth(\@AdjacencyMatrix): Calculates the girth of a graph. Quite slow, and 
# not actually used. Probably good for something.
sub girth {
	my $matref = shift;
	my @Matrix = @$matref;
	my $best = $numplayers + 1;

	my $i;
	for $i (1..$numplayers-2) {
		my @distList;
		my @span;
		$span[$i] = 1;
		my $depth = 1;
		push @distList, $i;
		while (($depth*2 <= $best) && ($best > 3)) {
			
			my @nextList;
			foreach $y (@distList) {
				my @nbrs;
				for $x (1..$numplayers) {
					push @nbrs, $x if ($Matrix[$x][$y]);
				}
				my $u;
				foreach $u (@nbrs) {
					if (!$span[$u]) {
						$span[$u] = 1;
						push @nextList, $u;
					}
					else {
						if (contains($u, @distList)) {
							$best = $depth*2 - 1; last;
						}
						if (contains($u, @nextList)) {
							$best = $depth*2;
						}
					}
				}
			}
			@distList = @nextList;
			$depth++;
		}
	}
	return $best;
}

# outputtargs(\@TargettingMatrix): Takes a big nasty targetting matrix and 
# prints a nice human-readable list.
sub outputtargs {
	my $matref = shift;
	my @Matrix = @$matref;
	for $x (1..$numplayers) {
		for $y (1..$numplayers) {
			print "$names[$x] -> $y\t" if ($Matrix[$x][$y] == 1);
		}
		print "\n";
	}
}

# adjacencylist(\@AdjacencyMatrix): Operates in O(n^2) and converts an 
# adjacency matrix into an adjacency list. Useful for some algorithms, 
# e.g. triangle counting
sub adjacencylist {
	my $matref = shift;
	my @Matrix = @$matref;
	my @adjlist;
	my ($i, $j);
	for $i (1..$numplayers) {
		for $j (1..$numplayers) {
			push @{$adjlist[$i]}, $j if $Matrix[$i][$j];
		}
	}
	return @adjlist;
}

# tricount(\@AdjacencyMatrix): Counts the 3-cliques in a graph in O(n^2)
# (assuming sparseness).
sub tricount {
	my $matref = shift;
	my @Matrix = @$matref;
	my @adjlist = adjacencylist(\@Matrix);
	my ($i, $j, $k);
	my $t = 0;
	for $i (1..$numplayers) {
		foreach $j (@{$adjlist[$i]}) {
			foreach $k (@{$adjlist[$i]}) {
				$t++ if $Matrix[$j][$k];
			}
		}
	}
	return $t/6;
}

# trilist(\@AdjacencyMatrix): Slightly modified version of tricount, returns an
# array of nodes involved in 3-cliques (with repetition)
sub trilist {
	my $matref = shift;
	my @Matrix = @$matref;
	my @adjlist = adjacencylist(\@Matrix);
	my ($i, $j, $k);
	my @tlist;
	for $i (1..$numplayers) {
		foreach $j (@{$adjlist[$i]}) {
			foreach $k (@{$adjlist[$i]}) {
				push @tlist, $i if $Matrix[$j][$k];
			}
		}
	}
	return @tlist;
}

sub playeridx {
	my $email = shift;
	for (1..$numplayers) {
		return $_ if ($emails[$_] eq $email);
	}
	return 0;
}	

sub swapplayers {
	my $i = shift;
	my $j = shift;
	($names[$i], $names[$j]) = ($names[$j], $names[$i]);
	($emails[$i], $emails[$j]) = ($emails[$j], $emails[$i]);
	($colls[$i], $colls[$j]) = ($colls[$j], $colls[$i]);
	($addrs[$i], $addrs[$j]) = ($addrs[$j], $addrs[$i]);
	($pseus[$i], $pseus[$j]) = ($pseus[$j], $pseus[$i]);
	($wwwws[$i], $wwwws[$j]) = ($wwwws[$j], $wwwws[$i]);
}
###############################################################################
###############################################################################
# Main code begins here. Well, a few lines down. But this is a comment, so ####
# you can treat it as beginning here. If that's what you're into. #############
###############################################################################
###############################################################################

# Turn on buffer flushing. Not actually necessary, but makes the output a bit
# easier on the eyes.
$| = 1;

open UINFO, ">testuinfo" or die;
open EVENTS, ">testevents" or die;

my $i = 1;
while (<>) {
	chomp;
	($names[$i], $emails[$i], $colls[$i], $addrs[$i], $pseus[$i], $wwwws[$i]) = split /\t/;
	$i++;
}
$numplayers = @names - 1;

# Blank everything. Necessary for some reason.
print "Voidicising...\n";
for $x (1..$numplayers) {
	for $y (1..$numplayers) {
		$TargMatrix[$x][$y] = 0;
	}
}

# Generate initial random targetting
print "Entargetulating...\n";
@TargMatrix = reallocate(\@TargMatrix);

# Use genetic trickery to remove 3-cliques
my $t = tricount(\@TargMatrix);
print "Detrianglatising.";
while ($t) {
	my @TrialMatrix = @TargMatrix;
	
	# Deallocate boring links with low probability
	for $x (1..$numplayers) {
		for $y (1..$x) {
			if (!(int(rand(5*$numplayers)))) {
				$TrialMatrix[$x][$y] = 0;
				$TrialMatrix[$y][$x] = 0;
			}
		}
	}
	
	# Deallocate interesting links with high probability
	my @tlist = trilist(\@TargMatrix);
	foreach $x (@tlist) {
		foreach $y (@tlist) {
			if (!(int(rand(2*$t)))) {
				$TrialMatrix[$x][$y] = 0;
				$TrialMatrix[$y][$x] = 0;
			}
		}
	}

	# Reallocate all removed links
	@TrialMatrix = reallocate(\@TrialMatrix);
	
	# Check for improvement, if there is one, keep it (or if no change)
	my $t2 = tricount(\@TrialMatrix);
	if ($t2 <= $t) { 
		print "." if $t > $t2;
		$t = $t2;
		@TargMatrix = @TrialMatrix;
	}
}
print "\n";

# Calculate pairwise distances. This bit is SLOW, but I can't see any way of 
# avoiding it, short of rethinking the entire seeding mechanism.
print "Floydulating...\n";
my @distance = floyd(\@TargMatrix);

# Place seeds. Seeds are placed at the greatest possible distance from all 
# existing seeds, using a heuristic we call "shininess". The shininess of a 
# location x is given by \Product_i (d(x, s_i) - 1) where s_i are the seed 
# locations, and d is the undirected distance function.

print "Interseedaging.";
my @seedlist;
push @seedlist, 1;
while (@seedlist < $numplayers) {
	my $maxshiny = -1000000;
	my $testseed;
	for $x (1..$numplayers) {
		my $shininess = 1;
		foreach $y (@seedlist) {
			$shininess *= $distance[$x][$y] - 1;
		}
		if (($shininess > $maxshiny) && !contains($x, @seedlist)) {
			$maxshiny = $shininess;
			$testseed = $x;
		}
	}
	last if !$maxshiny;
	push @seedlist, $testseed;
	print ".";
}
print "\n";
#outputtargs(\@TargMatrix);
#print "Seeds: $#seedlist\n";
#$, = ", ";
#print @seedlist;
#print "\n";
print "Input seed:\n";
$x = 0;
while (<STDIN>) {
	chomp;
	last if ($_ eq "");
	my $i = playeridx($_);
	print "Got $names[$i]\n";
	swapplayers($i,$seedlist[$x]);
	$x++;
	print "Input seed:\n";
}
for $x (1..$numplayers) {
	my $id = $x * 2 + 77;
	my @tid;
	print UINFO "*p$id\t(NoPassword)\t$names[$x]\t$colls[$x]\t$wwwws[$x]\t$addrs[$x]\t$emails[$x]\t4104\t$NOKES3\t0\t0\t0\t0\t1\t-\tNo comment\t$pseus[$x]\n";
	for $y (1..$numplayers) {
		push @tid, $y * 2 + 77 if $TargMatrix[$x][$y] == 1;
	}
	print EVENTS "*s$x\t$NOKES1\t103020\tp$id\t$NOKES2\t$pseus[$x]\tTp$id:p$tid[0]\tTp$id:p$tid[1]\tTp$id:p$tid[2]\t-\n";
}
