/**
 * Class to conduct arbitrary scoring operations on game data.
 * In the constructor we recieve a string which we then render 
 * into a machine conducable form, and then execute for a given 
 * assassin when execute is called. This algorithm is recursive 
 * if neccessary.
*/

package org.ucam.srcf.assassins.strawberry;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.ListIterator;
import java.util.Stack;

import org.ucam.srcf.assassins.AssassinViewer;
import org.ucam.srcf.assassins.Debug;
import org.ucam.srcf.assassins.exceptions.ConfigFileException;

class Scorer {
	/** Compares to operator <b>Class</b> objects, in reverse.*/
	private class op_Rcomp implements Comparator{
		private int op1_p;
		private int op2_p;
		
		/**
		 * Compare op1 with op2 wrt precedence value.
		 * For our reversed purposes lower is better.
		 * @return	neative if op1 less than op2, 0 if equal, 
		 *				positive if op1 greater than op2.
		 */
		public int compare(Object op1, Object op2) {
			try {
				op1_p=((Integer) ((Class) op1).getDeclaredMethod("getPrecedence",null).invoke(null,null)).intValue();
				op2_p=((Integer) ((Class) op2).getDeclaredMethod("getPrecedence",null).invoke(null,null)).intValue();
			} catch (Exception ex) {
				Debug.print(Scorer.this, "Eek! asked to compare non-Operator object(s?), claiming equal! ("+ex+")");
				return 0;
			}
			if(Debug.debug) Debug.print(Scorer.this, "Comparing "+op1+" and "+op2+" hence returning "+(op2_p-op1_p));
			return op2_p - op1_p;
		}
	}

	private ArrayList ops=new ArrayList();
	int start; /* declared here to avoid scoping conflicts*/
	/** List of all possible operations we can conduct */
	private Class[] possibleOperations = {
			Addition.class,
			Multiplication.class,
			Subtraction.class,
			Division.class,
			Power.class
			};

	/**
	 * Takes a string and sets itself up to carry out the specifyed ops.
	 * 
	 * We process according to BODMAS:
	 * B		Brackets
	 * O		Operations
	 * D		Divsion
	 * M		Multipication
	 * A		Addition
	 * S		Subtraction
	 *
	 * @param eqn String specifying equation to conduct
	 * @param offset Offset from beginning of string for error messages
	 */
	Scorer(String eqn, int offset) 
		throws ConfigFileException
	{
		int i=0;		/*counts our position in eqn*/
		Stack bStack=new Stack();		/* keeps track of how many 
													brackets we have open*/
		char c;		/*the current char we are examining*/
		ListIterator j;	/*Iterator for stepping thorugh various lists*/
		int k;	/*iterator*/
			
		/* Process the string into objects, we do the order
		 * of operations as part of the execute method */
		if(Debug.debug) {
			Debug.print(this, "Processing \""+eqn+"\" from offset "+offset);
		}
string_loop:
		while(i<eqn.length()) {
			Debug.print(this, "i:"+i);
			switch(c=eqn.charAt(i++)) {
				/* ignore white space*/
				case ' ':
					continue string_loop;

				/*handle brackets*/
				case ')':
					throw new ConfigFileException("Unmatched bracket in scoring equation at character "+(i+offset-1));
				case '(':
					start=i;
					bStack.push(new Character('('));
					while(!bStack.empty()) {
						switch(c=eqn.charAt(i++)) {
							case '(':
								bStack.push(new Character(c));
								break;
							case ')':
								bStack.pop();
								break;
						}
					}
					/*Now recurse on the string contained*/
					ops.add(new Scorer(eqn.substring(start,i-1),start));
					break;

				/*handle variables*/
				case '$':
					/*single letter or phrase in {}'s - not stackable!*/
					if( '{' == (c=eqn.charAt(i++)) ) {
						for(start=i; (c=eqn.charAt(i++))!='}'; );
						ops.add(new Variable(eqn.substring(start,i-1)));
					}
					else
						ops.add(new Variable(c));
					break;

				/*handle everything else*/
				default:
					/* Is it a number constant? */
					if( (c>='0' && c<='9') || c=='.') {
						/* go through to end of nuymbers */
						for(start=i-1;  i<eqn.length() && Character.isDigit(c=eqn.charAt(i)) ; i++);
						ops.add(new Constant(eqn.substring(start,i)));
						continue string_loop;
					}
					/* Go through individual operator checking if this 
					 *		is one of them*/
					for(k=0;k<possibleOperations.length;k++) {
						Class current_op=possibleOperations[k];
						/* the next line just calls the class's getCharacter mehod and compares it to c */
						try {
							Debug.print(this, "Testing :"+ current_op.getMethod("getCharacter",null).invoke(current_op,null));
							if(c == ((Character) current_op.getMethod("getCharacter",null).invoke(null,null)).charValue()) {
								ops.add(current_op.newInstance());
								if(Debug.debug) Debug.print(this, c+" operator instantiated: "+current_op);
								continue string_loop;
							}
						} catch (Exception ex) {
							/* Something wrong - couldn't instantiate a new 
							 * instance */
							Debug.print(this, "Eek! Couldn't instantiate an Operator class ("+c+"). This might hurt - "+ex);
						}
					}
					/*character not otherwise handled*/
					throw new ConfigFileException("Bad scoring equation at character "+(i+offset-1)+": " + c);
			}
		}
		
		/*Now sanity check it - avoid double operations*/
		j=ops.listIterator();
		boolean lastOp=true;
		while(j.hasNext()) {
			Object myItem=j.next();
			if(myItem instanceof Operator) {
				if(lastOp)
					throw new ConfigFileException("Bad scoring equation, wo consecutive operations!");
				lastOp=true;
			}
			else
				lastOp=false;
		}

		/*Finally sort the possibleOps into the correct precedence order*/
		Arrays.sort(possibleOperations, new op_Rcomp());
		if(Debug.debug) {
			Debug.print(this, "Order of operations:");
			for(i=0; i<possibleOperations.length; i++) {
				Debug.print(this, possibleOperations[i].toString());
			}
		}
	}

	/**
	 * Calculates the score.
	 * Carries out operations from opList accodring to BODMAS
	 *
	 * @param a assassin to score.
	 */
	public double execute(AssassinViewer killer, AssassinViewer victim, VarTracker vt) {
		/* First we create a List which we intend to make into one item
		 * Note that the objects in the list are _not_ copied, however we
		 * don't replace them per se so we should be fine.
		 * */
		ArrayList working=(ArrayList) ops.clone();
		
		/*Now we go through and put in the variables and recursive answers*/
		if(Debug.debug) Debug.print(this, "Evaluating variables and scorers");
		ListIterator i=working.listIterator();
		while(i.hasNext()) {
			Object current=i.next();
			if(current instanceof Scorer)
				i.set(new Double(((Scorer) current).execute(killer, victim, vt)));
			else if(current instanceof Variable)
				i.set(new Double(((Variable) current).evaluate(killer, victim, vt)));
			else if(current instanceof Constant)
				i.set(new Double(((Constant) current).evaluate()));
		}
		if(Debug.debug) Debug.print(this, "...done");

		/*Now we proceed to conduct the operations in the right order, 
		 * assuming sanity checks in constructor worked */
		if(Debug.debug) Debug.print(this, "Conducting operations");
		for(int j=0; j<possibleOperations.length; j++) {
			for(i=working.listIterator(); i.hasNext(); ) {
				Object current_obj=i.next();
				if(!(current_obj instanceof Operator))
					continue;
				Operator current=(Operator) current_obj;
				if(possibleOperations[j].isInstance(current)) {
					/*fancy footwork to delete stuff*/
					i.remove();
					double lval=((Double) i.previous()).doubleValue();
					i.remove();
					double rval=((Double) i.next()).doubleValue();
					i.set(new Double(current.evaluate(lval, rval)));
				}
			}
		}
		if(Debug.debug) Debug.print(this, "...done");

		/*we should now only have one item in the list now*/
		/*TODO sanity check for the above*/
		return ((Double) working.get(0)).doubleValue();
	}
	
}
